import os
from uuid import uuid4

from policyweaver.models.common import (
    SourceMap
)

class Configuration:
    @staticmethod
    def configure_environment(config:SourceMap):
        if not config.correlation_id:
            config.correlation_id = str(uuid4())

        os.environ['CORRELATION_ID'] = config.correlation_id


from policyweaver.models.common import CommonBaseModel, CommonBaseEnum
from typing import List, Optional
from pydantic import Field


class FabricPolicyAccessType(str, CommonBaseEnum):
    EXECUTE = "Execute"
    EXPLORE = "Explore"
    READ = "Read"
    READ_ALL = "ReadAll"
    RESHARE = "Reshare"
    WRITE = "Write"


class FabricMemberObjectType(str, CommonBaseEnum):
    GROUP = "Group"
    MANAGED_IDENTITY = "ManagedIdentity"
    SERVICE_PRINCIPAL = "ServicePrincipal"
    USER = "User"


class PolicyEffectType(str, CommonBaseEnum):
    PERMIT = "Permit"


class PolicyAttributeType(str, CommonBaseEnum):
    ACTION = "Action"
    PATH = "Path"


class EntraMember(CommonBaseModel):
    object_id: Optional[str] = Field(alias="objectId", default=None)
    object_type: Optional[FabricMemberObjectType] = Field(
        alias="objectType", default=None
    )
    tenant_id: Optional[str] = Field(alias="tenantId", default=None)


class PolicyMember(CommonBaseModel):
    source_path: Optional[str] = Field(alias="sourcePath", default=None)
    item_access: Optional[List[FabricPolicyAccessType]] = Field(
        alias="itemAccess", default=None
    )


class PolicyMembers(CommonBaseModel):
    fabric_members: Optional[List[PolicyMember]] = Field(
        alias="fabricItemMembers", default=None
    )
    entra_members: Optional[List[EntraMember]] = Field(
        alias="microsoftEntraMembers", default=None
    )


class PolicyPermissionScope(CommonBaseModel):
    attribute_name: Optional[PolicyAttributeType] = Field(
        alias="attributeName", default=None
    )
    attribute_value_included_in: Optional[List[str]] = Field(
        alias="attributeValueIncludedIn", default=None
    )


class PolicyDecisionRule(CommonBaseModel):
    effect: Optional[PolicyEffectType] = Field(alias="effect", default=None)
    permission: Optional[List[PolicyPermissionScope]] = Field(
        alias="permission", default=None
    )


class DataAccessPolicy(CommonBaseModel):
    id: Optional[str] = Field(alias="id", default=None)
    name: Optional[str] = Field(alias="name", default=None)
    decision_rules: Optional[List[PolicyDecisionRule]] = Field(
        alias="decisionRules", default=None
    )
    members: Optional[PolicyMembers] = Field(alias="members", default=None)



import requests


class RestAPIProxy:
    def __init__(self, base_url, headers=None):
        self.base_url = base_url
        self.headers = headers

    def get(self, endpoint, params=None, headers=None):
        if not headers:
            headers = self.headers

        response = requests.get(
            f"{self.base_url}/{endpoint}", params=params, headers=headers
        )
        return self._handle_response(response)

    def post(self, endpoint, data=None, json=None, files=None, headers=None):
        if not headers:
            headers = self.headers

        response = requests.post(
            f"{self.base_url}/{endpoint}",
            data=data,
            json=json,
            files=files,
            headers=headers,
        )
        return self._handle_response(response)

    def put(self, endpoint, data=None, json=None, headers=None):
        if not headers:
            headers = self.headers

        response = requests.put(
            f"{self.base_url}/{endpoint}", data=data, json=json, headers=headers
        )
        return self._handle_response(response)

    def delete(self, endpoint, headers=None):
        if not headers:
            headers = self.headers

        response = requests.delete(f"{self.base_url}/{endpoint}", headers=headers)
        return self._handle_response(response)

    def _handle_response(self, response):
        if response.status_code in (200, 201, 202):
            return response
        else:
            response.raise_for_status()



from typing import Optional, List, Any
from enum import Enum
from pydantic import BaseModel, Field, ConfigDict

import os
import re
import yaml

class PolicyWeaverError(Exception):
    pass

class CommonBaseModel(BaseModel):
    model_config = ConfigDict(
        populate_by_name=True,
        use_enum_values=True,
        arbitrary_types_allowed=True,
        exclude_unset=True,
        exclude_none=True,
    )

    def model_dump(self, **kwargs) -> dict[str, Any]:
        return super().model_dump(by_alias=True, **kwargs)

    def model_dump_json(self, **kwargs) -> dict[str, Any]:
        return super().model_dump_json(by_alias=True, **kwargs)

    def __getattr__(self, item):
        for field, meta in self.model_fields.items():
            if meta.alias == item:
                return getattr(self, field)
        return super().__getattr__(item)

    def _get_alias(self, item_name):
        for field, meta in self.model_fields.items():
            if field == item_name:
                return meta.alias

        return None


class CommonBaseEnum(Enum):
    def __str__(self):
        return str(self.value)


class IamType(str, CommonBaseEnum):
    USER = "USER"
    GROUP = "GROUP"


class PermissionType(str, CommonBaseEnum):
    SELECT = "SELECT"


class PermissionState(str, CommonBaseEnum):
    GRANT = "GRANT"


class PolicyWeaverConnectorType(str, CommonBaseEnum):
    UNITY_CATALOG = "UNITY_CATALOG"
    SNOWFLAKE = "SNOWFLAKE"
    BIGQUERY = "BIGQUERY"


class SourceSchema(CommonBaseModel):
    name: Optional[str] = Field(alias="name", default=None)
    tables: Optional[List[str]] = Field(alias="tables", default=None)


class CatalogItem(CommonBaseModel):
    catalog: Optional[str] = Field(alias="catalog", default=None)
    catalog_schema: Optional[str] = Field(alias="catalog_schema", default=None)
    table: Optional[str] = Field(alias="table", default=None)


class Source(CommonBaseModel):
    name: Optional[str] = Field(alias="name", default=None)
    schemas: Optional[List[SourceSchema]] = Field(alias="schemas", default=None)

    def get_schema_list(self) -> List[str]:
        if not self.schemas:
            return None

        return [s.name for s in self.schemas]


class PermissionObject(CommonBaseModel):
    id: Optional[str] = Field(alias="id", default=None)
    type: Optional[IamType] = Field(alias="type", default=None)


class Permission(CommonBaseModel):
    name: Optional[str] = Field(alias="name", default=None)
    state: Optional[str] = Field(alias="state", default=None)
    objects: Optional[List[PermissionObject]] = Field(alias="objects", default=None)


class Policy(CatalogItem):
    permissions: Optional[List[Permission]] = Field(alias="permissions", default=None)


class PolicyExport(CommonBaseModel):
    source: Optional[Source] = Field(alias="source", default=None)
    type: Optional[PolicyWeaverConnectorType] = Field(alias="type", default=None)
    policies: Optional[List[Policy]] = Field(alias="policies", default=None)


class FabricConfig(CommonBaseModel):
    tenant_id: Optional[str] = Field(alias="tenant_id", default=None)
    workspace_id: Optional[str] = Field(alias="workspace_id", default=None)
    workspace_name: Optional[str] = Field(alias="workspace_name", default=None)
    lakehouse_id: Optional[str] = Field(alias="lakehouse_id", default=None)
    lakehouse_name: Optional[str] = Field(alias="lakehouse_name", default=None)
    use_lakehouse_schema: Optional[bool] = Field(alias="use_lakehouse_schema", default=None)

class ServicePrincipalConfig(CommonBaseModel):
    tenant_id: Optional[str] = Field(alias="tenant_id", default=None)
    client_id: Optional[str] = Field(alias="client_id", default=None)
    client_secret: Optional[str] = Field(alias="client_secret", default=None)


class SourceMapItem(CatalogItem):
    lakehouse_table_name: Optional[str] = Field(
        alias="lakehouse_table_name", default=None
    )

class SourceMap(CommonBaseModel):
    application_name: Optional[str] = Field(alias="application_name", default="POLICY_WEAVER")
    correlation_id: Optional[str] = Field(alias="correlation_id", default=None)
    type: Optional[PolicyWeaverConnectorType] = Field(alias="type", default=None)
    source: Optional[Source] = Field(alias="source", default=None)
    fabric: Optional[FabricConfig] = Field(alias="fabric", default=None)
    service_principal: Optional[ServicePrincipalConfig] = Field(
        alias="service_principal", default=None
    )
    mapped_items: Optional[List[SourceMapItem]] = Field(
        alias="mapped_items", default=None
    )

    _default_paths = ['./settings.yaml']

    @classmethod
    def from_yaml(cls, path:str=None) -> 'SourceMap':
        paths = [path] if path else cls._default_paths.default
            
        for p in paths:
            if os.path.exists(p):
                with open(p, 'r', encoding="utf-8") as f:
                    data = yaml.safe_load(f)
                return cls(**data)
        
        raise PolicyWeaverError("Policy Sync settings file not found")

    def __save_to_first_writable_path__(self, path:str=None) -> None:
        paths = [path] if path else self._default_paths

        for p in paths:
            try:
                with open(p, 'w', encoding="utf-8") as f:
                    yaml.safe_dump(self.model_dump(exclude_none=True, exclude_unset=True), f)
                print(f"Settings saved to {p}")
                return
            except IOError:
                print(f"Unable to write to {p}")
        raise IOError(f"None of the paths in {paths} are writable.")

    def to_yaml(self, path:str=None) -> None:
        self.__save_to_first_writable_path__(path)

class Utils:
    @staticmethod
    def is_email(email):
        pattern = r"^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$"
        return re.match(pattern, email)


from policyweaver.models.common import (
    CommonBaseModel,
    CommonBaseEnum,
    IamType,
    SourceMap
)

from pydantic import Field
from typing import Optional, List, Dict

class UnityCatalogPrincipalType(str, CommonBaseEnum):
    USER = "USER"
    GROUP = "GROUP"

class DependencyMap(CommonBaseModel):
    catalog: Optional[str] = Field(alias="catalog", default=None)
    catalog_schema: Optional[str] = Field(alias="schema", default=None)
    table: Optional[str] = Field(alias="table", default=None)
    privileges: Optional[List[str]] = Field(alias="privileges", default=[])
    catalog_prerequisites: Optional[bool] = Field(alias="catalog_prerequisites", default=False)
    schema_prerequisites: Optional[bool] = Field(alias="schema_prerequisites", default=False)
    read_permissions: Optional[bool] = Field(alias="read_permissions", default=False)

    @property
    def key(self) -> str:
        s = f".{self.catalog_schema}" if self.catalog_schema else ""
        t = f".{self.table}" if self.table else ""

        return f"{self.catalog}{s}{t}".lower()

class PrivilegeSnapshot(CommonBaseModel):
    principal: Optional[str] = Field(alias="principal", default=None)
    type: Optional[UnityCatalogPrincipalType] = Field(alias="type", default=None)
    maps: Optional[Dict[str, DependencyMap]] = Field(alias="maps", default={})    
    group_membership: Optional[List[str]] = Field(alias="group_membership", default=[])

class Privilege(CommonBaseModel):
    principal: Optional[str] = Field(alias="principal", default=None)
    privileges: Optional[List[str]] = Field(alias="privileges", default=None)


class BaseObject(CommonBaseModel):
    id: Optional[str] = Field(alias="id", default=None)
    name: Optional[str] = Field(alias="name", default=None)


class PrivilegedObject(BaseObject):
    privileges: Optional[List[Privilege]] = Field(alias="privileges", default=None)


class FunctionMap(BaseObject):
    columns: Optional[List[str]] = Field(alias="column", default=None)


class Function(PrivilegedObject):
    sql: Optional[str] = Field(alias="sql", default=None)


class Table(PrivilegedObject):
    column_masks: Optional[List[FunctionMap]] = Field(
        alias="column_masks", default=None
    )
    row_filter: Optional[FunctionMap] = Field(alias="row_filter", default=None)


class Schema(PrivilegedObject):
    tables: Optional[List[Table]] = Field(alias="table", default=None)
    mask_functions: Optional[List[Function]] = Field(
        alias="mask_functions", default=None
    )

class Catalog(PrivilegedObject):
    schemas: Optional[List[Schema]] = Field(alias="schemas", default=None)


class WorkspaceUser(BaseObject):
    email: Optional[str] = Field(alias="email", default=None)


class WorkspaceGroupMember(BaseObject):
    type: Optional[IamType] = Field(alias="type", default=None)


class WorkspaceGroup(BaseObject):
    members: Optional[List[WorkspaceGroupMember]] = Field(alias="members", default=None)


class Workspace(BaseObject):
    catalog: Optional[Catalog] = Field(alias="catalog", default=None)
    users: Optional[List[WorkspaceUser]] = Field(alias="users", default=None)
    groups: Optional[List[WorkspaceGroup]] = Field(alias="groups", default=None)

    def lookup_user_by_id(self, id: str) -> WorkspaceUser:
        user = list(filter(lambda u: u.id == id, self.users))

        if user:
            return user[0]

        return None

    def lookup_user_by_email(self, email: str) -> WorkspaceUser:
        user = list(filter(lambda u: u.email == email, self.users))

        if user:
            return user[0]

        return None
    
    def lookup_group_by_name(self, name: str) -> WorkspaceUser:
        group = list(filter(lambda g: g.name == name, self.groups))

        if group:
            return group[0]
        
        return None
    
    def lookup_object_id(self, principal:str, type:IamType) -> str:
        if type == IamType.USER:
            u = self.lookup_user_by_email(principal)
            return u.id if u else None
        else:
            g = self.lookup_group_by_name(principal)
            return g.id if g else None
    
    def get_user_groups(self, object_id:str) -> List[str]:
        membership = []
        
        for g in self.groups:
            membership = self.__extend_with_dedup__(membership, 
                                                    self.__flatten_group__(g.name, object_id))                  

        return membership
    
    def __extend_with_dedup__(self, src, new):
        if not src or len(src) == 0:
            return new

        if not new or len(new) == 0:
            return src

        s = set(src)
        s.update(new)

        return list(s)
    
    def __flatten_group__(self, name: str, id: str) -> List[str]:
        group = self.lookup_group_by_name(name)
        membership = []

        if group:
            for m in group.members:
                if m.type == IamType.USER and m.id == id:
                    if m.name not in membership:
                        membership.append(group.name)
                elif m.type == IamType.GROUP:
                    membership = self.__extend_with_dedup__(
                        membership, self.__flatten_group__(m.name, id)
                    )

        return membership

class DatabricksSourceMap(SourceMap):
    workspace_url: Optional[str] = Field(alias="workspace_url", default=None)